#!/usr/bin/env python3

from .Input_Parsers.Clingo_Parser.clingo_parser import parse_clingo_output
import pandas as pd
import numpy as np

import os
from .pwe_helper import preprocess_clingo_output


def parse_solution(fname, reasoner='clingo'):

    parser_to_use = None
    if reasoner == 'clingo':
        parser_to_use = parse_clingo_output
    else:
        print("Unrecognized reasoner selected")
        exit(1)

    dfs, relations, pws = parser_to_use(fname)
    return dfs, relations, pws


def load_worlds(clingo_output: list, reasoner='clingo', preprocessed: bool=True):

    if not preprocessed:
        clingo_output = preprocess_clingo_output(clingo_output)

    # TODO Add functionality to generate random file name
    dummy_fname = 'sjbcbshlpowieiohbcjhsbnckibubkjcnaiuhwyegvjcbwscuawhbnckbuveyrb.txt'
    with open(dummy_fname, 'w') as f:
        f.write('\n'.join(clingo_output))
    dfs, relations, pws = parse_solution(dummy_fname, reasoner)
    os.remove(dummy_fname)
    return dfs, relations, pws



#!/usr/bin/env python3

import pandas as pd
import numpy as np
import os

import matplotlib.pyplot as plt
from sklearn.cluster import DBSCAN
from scipy.spatial.distance import squareform
from scipy.cluster.hierarchy import dendrogram, linkage
import networkx as nx
from sklearn.manifold import MDS
# from sklearn.decomposition import PCA


def dbscan_clustering(dist_matrix, save_to_file=None):

    fig, ax = plt.subplots()

    db = DBSCAN(metric='precomputed', eps=0.5, min_samples=1)
    labels = db.fit_predict(dist_matrix)

    core_samples_mask = np.zeros_like(db.labels_, dtype=bool)
    core_samples_mask[db.core_sample_indices_] = True
    labels = db.labels_

    # Number of clusters in labels, ignoring noise if present.
    n_clusters_ = len(set(labels)) - (1 if -1 in labels else 0)
    unique_labels = set(labels)
    colors = [plt.cm.Spectral(each)
              for each in np.linspace(0, 1, len(unique_labels))]

    # dist_matrix = PCA(n_components = 2).fit_transform(dist_matrix)

    for k, col in zip(unique_labels, colors):
        if k == -1:
            # Black used for noise.
            col = [0, 0, 0, 1]

        class_member_mask = (labels == k)
        xy = dist_matrix[class_member_mask & core_samples_mask]
        ax.plot(xy[:, 0], xy[:, 1], 'o', markerfacecolor=tuple(col), markeredgecolor='k', markersize=14)

        xy = dist_matrix[class_member_mask & ~core_samples_mask]
        ax.plot(xy[:, 0], xy[:, 1], 'o', markerfacecolor=tuple(col), markeredgecolor='k', markersize=6)

    ax.title('Estimated number of clusters: %d' % n_clusters_)
    if save_to_file is not None:
        fig.savefig(save_to_file)
    return fig, labels


def linkage_dendrogram(dist_matrix, save_to_folder=None):

    X = squareform(dist_matrix)
    dendrogram_size = (max(25, int(np.sqrt(2 * len(X)) / 10)), 10)
    figs = []
    for dist_type in ['single', 'complete', 'average', 'weighted']:
        fig, ax = plt.subplots(figsize=dendrogram_size)
        linkage_matrix = linkage(X, dist_type)
        dendrogram(linkage_matrix, labels=[str(i) for i in range(len(dist_matrix))], show_leaf_counts=True, ax=ax)
        ax.title("Dendrogram ({})".format(dist_type))
        if save_to_folder is not None:
            fig.savefig(os.path.join(save_to_folder, '{}_dendrogram.png'.format(dist_type)))
        figs.append(fig)
    return figs


def mds_graph_2(pws, A, scale_down_factor, save_to_file=True):

    dt = [('len', float)]
    A = A * len(A) / scale_down_factor
    A = A.view(dt)
    G = nx.from_numpy_matrix(A)
    G = nx.relabel_nodes(G,
                         dict(list(zip(list(range(len(G.nodes()))), ['pw-{}'.format(i) for i in range(0, len(pws))]))))

    G = nx.drawing.nx_agraph.to_agraph(G)

    G.node_attr.update(color="red", style="filled")
    G.edge_attr.update(color=None, width="0.1")
    # G.edge_attr.update(color="blue", width="0.1")

    if save_to_file is not None:
        G.draw(save_to_file, format='png', prog='neato')

    return G


def mds_sklearn(A, save_to_file=None):

    fig, ax = plt.subplots()
    mds = MDS(2, dissimilarity="precomputed")
    mds.fit(A)
    x = mds.embedding_[:, 0]
    y = mds.embedding_[:, 1]
    ax.scatter(x, y)
    if save_to_file is not None:
        fig.savefig(save_to_file)
    return fig



#!/usr/bin/env python3

from PW_explorer.pwe_helper import rel_id_from_rel_name, mkdir_p
from graphviz import Digraph
import os
from collections import defaultdict

styles = {
    'graph': {
        'label': 'Taxonomy Alignment',
        'fontsize': '16',
        'fontcolor': 'black',
        'bgcolor': '#ffffff',
        'rankdir': 'LR',
    },
    'node_styles': {
        'node_equal': {
            'shape': 'box',
            'style': 'filled,rounded',
            'fontname': 'helvetica',
            'fillcolor': '#EEEEEE',
            'fontcolor': 'black',
        },
    },
    'edge_styles': {
        'overlap_edge': {
            'arrowhead': 'none',
            'style': 'dotted',
            'constraint': 'false',
            'penwidth': '1',
            'color': '#ce2118',
        },
        'proper_part_edge': {
            'arrowhead': 'normal',
            'style': 'solid',
            'color': 'black',
            'constraint': 'true',
            'penwidth': '1',
        },
    }
}


def remove_quotes(x):
    if x[0] == '"':
        x = x[1:]
    if x[-1] == '"':
        x = x[:-1]
    return x


def generate_node_style(fillcolor, fontcolor='black', shape='box', style='filled', fontname='helvetica'):
    return dict(fillcolor=fillcolor, fontcolor=fontcolor, shape=shape, style=style, fontname=fontname)


def find_set(ns, name):
    for i, ns_ in enumerate(ns):
        if name in ns_:
            return i


def visualize(**kwargs):

    kwargs = defaultdict(lambda: None, kwargs)
    dfs = kwargs['dfs']
    pws = kwargs['pws']
    relations = kwargs['relations']
    save_to_folder = kwargs['save_to_folder']
    graphs = []
    for pw_id in range(1, len(pws) + 1):
        graph = Digraph(graph_attr=styles['graph'])
        nodes_sets = []
        # Get the regions/concepts from the 'u' df
        df = dfs[rel_id_from_rel_name('u_1', relations)]
        df = df[df.pw == pw_id]
        for idx, row in df.iterrows():
            nodes_sets.append(set([row['x1']]))

        df = dfs[rel_id_from_rel_name('eq_2', relations)]
        df = df[df.pw == pw_id]

        for idx, row in df.iterrows():
            if row['x1'] == row['x2']:
                continue
            for i, set1 in enumerate(nodes_sets):
                if row['x1'] in set1:
                    for j, set2 in enumerate(nodes_sets):
                        if row['x2'] in set2:
                            if i != j:
                                nodes_sets[i] = nodes_sets[i].union(nodes_sets[j])
                                nodes_sets.pop(j)
                                break
                    break

        nodes_sets = list(map(list, nodes_sets))
        NODE_COLORS = ['#CCFFCC', '#FFFFCC', '#f4bf42', '#6346d6']
        NODE_COLORS_USED = 0
        for node_set in nodes_sets:
            node_set = list(map(remove_quotes, node_set))
            if len(node_set) == 1:
                tax = node_set[0].split('_')[0]
                if tax not in styles['node_styles']:
                    new_node_style = generate_node_style(NODE_COLORS[NODE_COLORS_USED])
                    NODE_COLORS_USED += 1
                    styles['node_styles'][tax] = new_node_style

                graph.node(node_set[0], _attributes=styles['node_styles'][tax])

            else:

                graph.node("\n".join(node_set), _attributes=styles['node_styles']['node_equal'])

        df = dfs[rel_id_from_rel_name('pp_2', relations)]
        df = df[df.pw == pw_id]

        pp_edges = []
        for idx, row in df.iterrows():
            node1 = list(map(remove_quotes, nodes_sets[find_set(nodes_sets, row['x1'])]))
            node2 = list(map(remove_quotes, nodes_sets[find_set(nodes_sets, row['x2'])]))
            if (node1, node2) not in pp_edges and (node2, node1) not in pp_edges:
                graph.edge("\n".join(node1), "\n".join(node2), _attributes=styles['edge_styles']['proper_part_edge'])
                pp_edges.append((node1, node2))

        df = dfs[rel_id_from_rel_name('po_2', relations)]
        df = df[df.pw == pw_id]

        overlap_edges = []
        for idx, row in df.iterrows():
            node1 = list(map(remove_quotes, nodes_sets[find_set(nodes_sets, row['x1'])]))
            node2 = list(map(remove_quotes, nodes_sets[find_set(nodes_sets, row['x2'])]))
            if (node1, node2) not in overlap_edges and (node2, node1) not in overlap_edges:
                graph.edge("\n".join(node1), "\n".join(node2), _attributes=styles['edge_styles']['overlap_edge'])
                overlap_edges.append((node1, node2))

        graphs.append(graph)

    if save_to_folder is not None:
        folder_name = os.path.join(save_to_folder, 'euler_visualization')
        mkdir_p(folder_name)
        for i, graph in enumerate(graphs):
            graph.render(filename='{}/pw-{}'.format(folder_name, i+1))

    return graphs



#!/usr/bin/env python3

from PW_explorer.pwe_helper import rel_id_from_rel_name, mkdir_p
from nxpd import draw
import networkx as nx
import os
from collections import defaultdict


def get_styles(project_name, pw_id):
    return {
        'graph': {
            'label': '{}_pw_{}'.format(str(project_name), str(pw_id)),
            'fontsize': '16',
            'fontcolor': 'black',
            'bgcolor': '#ffffff',
            'rankdir': 'RL',
        },
        'node_styles': {
            'node_equal': {
                'shape': 'box',
                'style': '"filled,rounded"',
                'fontname': 'helvetica',
                'fillcolor': '#EEEEEE',
                'fontcolor': 'black',
            },
        },
        'edge_styles': {
            'overlap_edge': {
                'arrowhead': 'none',
                'style': 'dotted',
                'constraint': 'false',
                'penwidth': '1',
                'color': '#ce2118',
            },
            'proper_part_edge': {
                'arrowhead': 'normal',
                'style': 'solid',
                'color': 'black',
                'constraint': 'true',
                'penwidth': '1',
            },
        }
    }

# Utility functions


def remove_quotes(x):
    if x[0] == '"':
        x = x[1:]
    if x[-1] == '"':
        x = x[:-1]
    return x


def generate_node_style(fillcolor, fontcolor='black', shape='box', style='filled', fontname='helvetica'):
    return dict(fillcolor=fillcolor, fontcolor=fontcolor, shape=shape, style=style, fontname=fontname)


# Disjoint Sets Utility Functions

def get_parent(node, djs):
    if djs[node] == node:
        return node
    parent = get_parent(djs[node], djs)
    djs[node] = parent
    return parent


def union(node1, node2, djs):
    node1 = get_parent(node1, djs)
    node2 = get_parent(node2, djs)
    if node1 != node2:
        djs[node2] = node1
    return node1


def setup_djs(djs):
    for child, parent in djs.items():
        djs[child] = get_parent(child, djs)

# Nodes merge function


def merge_nodes(G,nodes, new_node, attr):
    G_ = G
    for i in range(1, len(nodes)):
        G_ = nx.contracted_nodes(G_, nodes[0], nodes[i], self_loops=False)  # Only works for two nodes at a time apparently
    G__ = nx.relabel.relabel_nodes(G_, {nodes[0] : new_node})
    G__.nodes[new_node].update(attr)
    return G__


# Visualization function

def visualize(**kwargs):

    kwargs = defaultdict(lambda: None, kwargs)
    dfs = kwargs['dfs']
    pws = kwargs['pws']
    relations = kwargs['relations']
    save_to_folder = kwargs['save_to_folder']
    project_name = kwargs['project_name']
    Gs = []

    for pw_id in range(1, len(pws) + 1):

        styles = get_styles(project_name, pw_id)
        G = nx.DiGraph(**styles['graph'])

        # Add all the units to the graph
        df = dfs[rel_id_from_rel_name('u_1', relations)]
        df = df[df.pw == pw_id]
        NODE_COLORS = ['#CCFFCC', '#FFFFCC', '#f4bf42', '#6346d6']
        NODE_COLORS_USED = 0

        for idx, row in df.iterrows():
            node_name = remove_quotes(row['x1'])
            tax = node_name.split('_')[0]
            if tax not in styles['node_styles']:
                new_node_style = generate_node_style(NODE_COLORS[NODE_COLORS_USED])
                NODE_COLORS_USED += 1
                styles['node_styles'][tax] = new_node_style
            G.add_node(node_name, **styles['node_styles'][tax])

        # Add proper part edges
        df = dfs[rel_id_from_rel_name('pp_2', relations)]
        df = df[df.pw == pw_id]
        for idx, row in df.iterrows():
            G.add_edge(remove_quotes(row['x1']), remove_quotes(row['x2']), **styles['edge_styles']['proper_part_edge'])

        # Remove the redundant edges i.e. edges that go to ancestors of a parent
        for node in G.nodes:
            pred = list(G.predecessors(node))
            succ = list(G.successors(node))
            # print('node: {}'.format(node))
            for pred_ in pred:
                # print('predecessor: {}'.format(pred_))
                for succ_ in succ:
                    # print('successor: {}'.format(succ_))
                    if G.has_edge(pred_, succ_):
                        #  print('removing edge {} {}'.format(pred_, succ_))
                        G.remove_edge(pred_, succ_)

        # Add partial overlap edges
        df = dfs[rel_id_from_rel_name('po_2', relations)]
        df = df[df.pw == pw_id]

        for idx, row in df.iterrows():
            n1 = remove_quotes(row['x1'])
            n2 = remove_quotes(row['x2'])
            if not G.has_edge(n2, n1):
                G.add_edge(n1, n2, **styles['edge_styles']['overlap_edge'])

        # Merge the equivalent nodes
        df = dfs[rel_id_from_rel_name('eq_2', relations)]
        df = df[df.pw==pw_id]

        # Find the equivalent sets
        nodes = list(G.nodes)
        dj_sets = dict(zip(nodes, nodes))
        for idx, row in df.iterrows():
            n1 = remove_quotes(row['x1'])
            n2 = remove_quotes(row['x2'])
            if n1 == n2:
                continue
            else:
                union(n1, n2, dj_sets)
        setup_djs(dj_sets)

        final_djs = {}
        for child, parent in dj_sets.items():
            if parent not in final_djs:
                final_djs[parent] = set([])
            final_djs[parent].add(child)
        final_djs = list(map(list, final_djs.values()))

        final_djs = list(filter(lambda x: len(x) > 1, final_djs))

        # Merge equivalent nodes
        for final_djs_ in final_djs:
            G = merge_nodes(G, final_djs_, '\n'.join(final_djs_), styles['node_styles']['node_equal'])

        # TODO Merged nodes may have redundant edges (direct edges to successors of parent)

        Gs.append(G)

    if save_to_folder is not None:
        folder_name = os.path.join(save_to_folder, 'euler_visualization')
        mkdir_p(folder_name)
        for i, G in enumerate(Gs):
            draw(G, format='gv', filename='{}/pw-{}'.format(folder_name, i+1))
            draw(G, format='pdf', filename='{}/pw-{}.pdf'.format(folder_name, i+1))

    return Gs











#!/usr/bin/env python3

import pandas as pd
import numpy as np
from .pwe_query import PWEQuery


def euler_complexity_analysis(relations, expected_pws, dfs, rl_id, col_name, pws_to_consider: list=None, do_print=True):

    if not pws_to_consider:
        pws_to_consider = [j for j in range(1, expected_pws + 1)]

    complexities = np.zeros(len(pws_to_consider))

    for i, pw in enumerate(pws_to_consider):
        complexities[i] = PWEQuery.freq(relations, expected_pws, dfs, rl_id, [col_name], ['"><"'], [pw], False)[1][0]

    if np.max(complexities) != np.min(complexities):
        complexities = (complexities - np.min(complexities)) / (np.max(complexities) - np.min(complexities))
    if do_print:
        paired_pw_compl = list(zip(pws_to_consider, complexities))
        paired_pw_compl = sorted(paired_pw_compl, key=lambda x: x[1], reverse=True)
        print('PWs:         ', str([x[0] for x in paired_pw_compl]))
        print('Complexities:', str([x[1].round(2) for x in paired_pw_compl]))

    return complexities



#!/usr/bin/env python3

import os
import subprocess as subprocess
from .pwe_helper import preprocess_clingo_output

def get_clingo_output(clingo_in_fnames: list, num_solutions: int=0):

    t = ['clingo', '-n {}'.format(num_solutions), '-Wnone']
    t.extend(clingo_in_fnames)
    process_ = subprocess.Popen(t, stdout=subprocess.PIPE)
    clingo_output = process_.communicate()[0]
    # clingo_output = subprocess.check_output()

    # print clingo_output
    cling_out_lines = clingo_output.splitlines()
    cling_out_lines = list(map(lambda x: str(x, 'utf-8'), cling_out_lines))
    # cling_out_lines = [l.decode('utf-8') for l in cling_out_lines]
    return cling_out_lines


def run_clingo(clingo_rules: list) -> list:

    dummy_fname = 'svjsihkankjbyerhoihsyvgjnclsdihcysbfhcbygweincbsydgibwyebcsygdyc.lp4'
    with open(dummy_fname, 'w') as f:
        f.write('\n'.join(clingo_rules))

    clingo_output = get_clingo_output([dummy_fname])
    os.remove(dummy_fname)
    return preprocess_clingo_output(clingo_output)




# Generated from Clingo.g4 by ANTLR 4.7.1
# encoding: utf-8
from antlr4 import *
from io import StringIO
from typing.io import TextIO
import sys

def serializedATN():
    with StringIO() as buf:
        buf.write("\3\u608b\ua72a\u8133\ub9ed\u417c\u3be7\u7786\u5964\3\17")
        buf.write("d\4\2\t\2\4\3\t\3\4\4\t\4\4\5\t\5\4\6\t\6\4\7\t\7\4\b")
        buf.write("\t\b\4\t\t\t\4\n\t\n\4\13\t\13\4\f\t\f\3\2\7\2\32\n\2")
        buf.write("\f\2\16\2\35\13\2\3\2\5\2 \n\2\3\2\3\2\3\3\3\3\3\3\7\3")
        buf.write("\'\n\3\f\3\16\3*\13\3\3\3\5\3-\n\3\3\3\5\3\60\n\3\3\4")
        buf.write("\6\4\63\n\4\r\4\16\4\64\3\4\3\4\3\4\3\4\3\5\3\5\3\6\3")
        buf.write("\6\5\6?\n\6\3\6\5\6B\n\6\3\6\3\6\3\6\3\6\3\7\3\7\3\7\3")
        buf.write("\7\5\7L\n\7\3\b\3\b\3\b\3\b\3\t\3\t\3\t\3\t\3\n\3\n\3")
        buf.write("\n\3\n\3\13\6\13[\n\13\r\13\16\13\\\3\f\6\f`\n\f\r\f\16")
        buf.write("\fa\3\f\2\2\r\2\4\6\b\n\f\16\20\22\24\26\2\3\5\2\5\6\b")
        buf.write("\b\16\16\2c\2\33\3\2\2\2\4#\3\2\2\2\6\62\3\2\2\2\b:\3")
        buf.write("\2\2\2\n<\3\2\2\2\fG\3\2\2\2\16M\3\2\2\2\20Q\3\2\2\2\22")
        buf.write("U\3\2\2\2\24Z\3\2\2\2\26_\3\2\2\2\30\32\5\4\3\2\31\30")
        buf.write("\3\2\2\2\32\35\3\2\2\2\33\31\3\2\2\2\33\34\3\2\2\2\34")
        buf.write("\37\3\2\2\2\35\33\3\2\2\2\36 \7\r\2\2\37\36\3\2\2\2\37")
        buf.write(" \3\2\2\2 !\3\2\2\2!\"\5\n\6\2\"\3\3\2\2\2#$\7\3\2\2$")
        buf.write("(\7\16\2\2%\'\5\6\4\2&%\3\2\2\2\'*\3\2\2\2(&\3\2\2\2(")
        buf.write(")\3\2\2\2),\3\2\2\2*(\3\2\2\2+-\7\4\2\2,+\3\2\2\2,-\3")
        buf.write("\2\2\2-/\3\2\2\2.\60\7\16\2\2/.\3\2\2\2/\60\3\2\2\2\60")
        buf.write("\5\3\2\2\2\61\63\7\16\2\2\62\61\3\2\2\2\63\64\3\2\2\2")
        buf.write("\64\62\3\2\2\2\64\65\3\2\2\2\65\66\3\2\2\2\66\67\7\5\2")
        buf.write("\2\678\5\b\5\289\7\6\2\29\7\3\2\2\2:;\7\16\2\2;\t\3\2")
        buf.write("\2\2<>\5\f\7\2=?\5\16\b\2>=\3\2\2\2>?\3\2\2\2?A\3\2\2")
        buf.write("\2@B\5\20\t\2A@\3\2\2\2AB\3\2\2\2BC\3\2\2\2CD\5\22\n\2")
        buf.write("DE\5\24\13\2EF\5\26\f\2F\13\3\2\2\2GH\7\7\2\2HI\7\b\2")
        buf.write("\2IK\7\16\2\2JL\7\t\2\2KJ\3\2\2\2KL\3\2\2\2L\r\3\2\2\2")
        buf.write("MN\7\n\2\2NO\7\b\2\2OP\7\16\2\2P\17\3\2\2\2QR\7\13\2\2")
        buf.write("RS\7\b\2\2ST\7\16\2\2T\21\3\2\2\2UV\7\f\2\2VW\7\b\2\2")
        buf.write("WX\7\16\2\2X\23\3\2\2\2Y[\t\2\2\2ZY\3\2\2\2[\\\3\2\2\2")
        buf.write("\\Z\3\2\2\2\\]\3\2\2\2]\25\3\2\2\2^`\t\2\2\2_^\3\2\2\2")
        buf.write("`a\3\2\2\2a_\3\2\2\2ab\3\2\2\2b\27\3\2\2\2\r\33\37(,/")
        buf.write("\64>AK\\a")
        return buf.getvalue()


class ClingoParser ( Parser ):

    grammarFileName = "Clingo.g4"

    atn = ATNDeserializer().deserialize(serializedATN())

    decisionsToDFA = [ DFA(ds, i) for i, ds in enumerate(atn.decisionToState) ]

    sharedContextCache = PredictionContextCache()

    literalNames = [ "<INVALID>", "'Answer:'", "'Optimization:'", "'('", 
                     "')'", "'Models'", "':'", "'+'", "'Optimum'", "'Optimization'", 
                     "'Calls'" ]

    symbolicNames = [ "<INVALID>", "<INVALID>", "<INVALID>", "<INVALID>", 
                      "<INVALID>", "<INVALID>", "<INVALID>", "<INVALID>", 
                      "<INVALID>", "<INVALID>", "<INVALID>", "OPTIMUM_FOUND", 
                      "TEXT", "WHITESPACE" ]

    RULE_clingoOutput = 0
    RULE_solution = 1
    RULE_actual_soln = 2
    RULE_custom_representation_soln = 3
    RULE_summary = 4
    RULE_models = 5
    RULE_optimum = 6
    RULE_optimization = 7
    RULE_calls = 8
    RULE_time = 9
    RULE_cpuTime = 10

    ruleNames =  [ "clingoOutput", "solution", "actual_soln", "custom_representation_soln", 
                   "summary", "models", "optimum", "optimization", "calls", 
                   "time", "cpuTime" ]

    EOF = Token.EOF
    T__0=1
    T__1=2
    T__2=3
    T__3=4
    T__4=5
    T__5=6
    T__6=7
    T__7=8
    T__8=9
    T__9=10
    OPTIMUM_FOUND=11
    TEXT=12
    WHITESPACE=13

    def __init__(self, input:TokenStream, output:TextIO = sys.stdout):
        super().__init__(input, output)
        self.checkVersion("4.7.1")
        self._interp = ParserATNSimulator(self, self.atn, self.decisionsToDFA, self.sharedContextCache)
        self._predicates = None



    class ClingoOutputContext(ParserRuleContext):

        def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
            super().__init__(parent, invokingState)
            self.parser = parser

        def summary(self):
            return self.getTypedRuleContext(ClingoParser.SummaryContext,0)


        def solution(self, i:int=None):
            if i is None:
                return self.getTypedRuleContexts(ClingoParser.SolutionContext)
            else:
                return self.getTypedRuleContext(ClingoParser.SolutionContext,i)


        def OPTIMUM_FOUND(self):
            return self.getToken(ClingoParser.OPTIMUM_FOUND, 0)

        def getRuleIndex(self):
            return ClingoParser.RULE_clingoOutput

        def enterRule(self, listener:ParseTreeListener):
            if hasattr( listener, "enterClingoOutput" ):
                listener.enterClingoOutput(self)

        def exitRule(self, listener:ParseTreeListener):
            if hasattr( listener, "exitClingoOutput" ):
                listener.exitClingoOutput(self)




    def clingoOutput(self):

        localctx = ClingoParser.ClingoOutputContext(self, self._ctx, self.state)
        self.enterRule(localctx, 0, self.RULE_clingoOutput)
        self._la = 0 # Token type
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 25
            self._errHandler.sync(self)
            _la = self._input.LA(1)
            while _la==ClingoParser.T__0:
                self.state = 22
                self.solution()
                self.state = 27
                self._errHandler.sync(self)
                _la = self._input.LA(1)

            self.state = 29
            self._errHandler.sync(self)
            _la = self._input.LA(1)
            if _la==ClingoParser.OPTIMUM_FOUND:
                self.state = 28
                self.match(ClingoParser.OPTIMUM_FOUND)


            self.state = 31
            self.summary()
        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx

    class SolutionContext(ParserRuleContext):

        def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
            super().__init__(parent, invokingState)
            self.parser = parser

        def TEXT(self, i:int=None):
            if i is None:
                return self.getTokens(ClingoParser.TEXT)
            else:
                return self.getToken(ClingoParser.TEXT, i)

        def actual_soln(self, i:int=None):
            if i is None:
                return self.getTypedRuleContexts(ClingoParser.Actual_solnContext)
            else:
                return self.getTypedRuleContext(ClingoParser.Actual_solnContext,i)


        def getRuleIndex(self):
            return ClingoParser.RULE_solution

        def enterRule(self, listener:ParseTreeListener):
            if hasattr( listener, "enterSolution" ):
                listener.enterSolution(self)

        def exitRule(self, listener:ParseTreeListener):
            if hasattr( listener, "exitSolution" ):
                listener.exitSolution(self)




    def solution(self):

        localctx = ClingoParser.SolutionContext(self, self._ctx, self.state)
        self.enterRule(localctx, 2, self.RULE_solution)
        self._la = 0 # Token type
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 33
            self.match(ClingoParser.T__0)
            self.state = 34
            self.match(ClingoParser.TEXT)
            self.state = 38
            self._errHandler.sync(self)
            _alt = self._interp.adaptivePredict(self._input,2,self._ctx)
            while _alt!=2 and _alt!=ATN.INVALID_ALT_NUMBER:
                if _alt==1:
                    self.state = 35
                    self.actual_soln() 
                self.state = 40
                self._errHandler.sync(self)
                _alt = self._interp.adaptivePredict(self._input,2,self._ctx)

            self.state = 42
            self._errHandler.sync(self)
            _la = self._input.LA(1)
            if _la==ClingoParser.T__1:
                self.state = 41
                self.match(ClingoParser.T__1)


            self.state = 45
            self._errHandler.sync(self)
            _la = self._input.LA(1)
            if _la==ClingoParser.TEXT:
                self.state = 44
                self.match(ClingoParser.TEXT)


        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx

    class Actual_solnContext(ParserRuleContext):

        def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
            super().__init__(parent, invokingState)
            self.parser = parser

        def custom_representation_soln(self):
            return self.getTypedRuleContext(ClingoParser.Custom_representation_solnContext,0)


        def TEXT(self, i:int=None):
            if i is None:
                return self.getTokens(ClingoParser.TEXT)
            else:
                return self.getToken(ClingoParser.TEXT, i)

        def getRuleIndex(self):
            return ClingoParser.RULE_actual_soln

        def enterRule(self, listener:ParseTreeListener):
            if hasattr( listener, "enterActual_soln" ):
                listener.enterActual_soln(self)

        def exitRule(self, listener:ParseTreeListener):
            if hasattr( listener, "exitActual_soln" ):
                listener.exitActual_soln(self)




    def actual_soln(self):

        localctx = ClingoParser.Actual_solnContext(self, self._ctx, self.state)
        self.enterRule(localctx, 4, self.RULE_actual_soln)
        self._la = 0 # Token type
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 48 
            self._errHandler.sync(self)
            _la = self._input.LA(1)
            while True:
                self.state = 47
                self.match(ClingoParser.TEXT)
                self.state = 50 
                self._errHandler.sync(self)
                _la = self._input.LA(1)
                if not (_la==ClingoParser.TEXT):
                    break

            self.state = 52
            self.match(ClingoParser.T__2)
            self.state = 53
            self.custom_representation_soln()
            self.state = 54
            self.match(ClingoParser.T__3)
        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx

    class Custom_representation_solnContext(ParserRuleContext):

        def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
            super().__init__(parent, invokingState)
            self.parser = parser

        def TEXT(self):
            return self.getToken(ClingoParser.TEXT, 0)

        def getRuleIndex(self):
            return ClingoParser.RULE_custom_representation_soln

        def enterRule(self, listener:ParseTreeListener):
            if hasattr( listener, "enterCustom_representation_soln" ):
                listener.enterCustom_representation_soln(self)

        def exitRule(self, listener:ParseTreeListener):
            if hasattr( listener, "exitCustom_representation_soln" ):
                listener.exitCustom_representation_soln(self)




    def custom_representation_soln(self):

        localctx = ClingoParser.Custom_representation_solnContext(self, self._ctx, self.state)
        self.enterRule(localctx, 6, self.RULE_custom_representation_soln)
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 56
            self.match(ClingoParser.TEXT)
        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx

    class SummaryContext(ParserRuleContext):

        def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
            super().__init__(parent, invokingState)
            self.parser = parser

        def models(self):
            return self.getTypedRuleContext(ClingoParser.ModelsContext,0)


        def calls(self):
            return self.getTypedRuleContext(ClingoParser.CallsContext,0)


        def time(self):
            return self.getTypedRuleContext(ClingoParser.TimeContext,0)


        def cpuTime(self):
            return self.getTypedRuleContext(ClingoParser.CpuTimeContext,0)


        def optimum(self):
            return self.getTypedRuleContext(ClingoParser.OptimumContext,0)


        def optimization(self):
            return self.getTypedRuleContext(ClingoParser.OptimizationContext,0)


        def getRuleIndex(self):
            return ClingoParser.RULE_summary

        def enterRule(self, listener:ParseTreeListener):
            if hasattr( listener, "enterSummary" ):
                listener.enterSummary(self)

        def exitRule(self, listener:ParseTreeListener):
            if hasattr( listener, "exitSummary" ):
                listener.exitSummary(self)




    def summary(self):

        localctx = ClingoParser.SummaryContext(self, self._ctx, self.state)
        self.enterRule(localctx, 8, self.RULE_summary)
        self._la = 0 # Token type
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 58
            self.models()
            self.state = 60
            self._errHandler.sync(self)
            _la = self._input.LA(1)
            if _la==ClingoParser.T__7:
                self.state = 59
                self.optimum()


            self.state = 63
            self._errHandler.sync(self)
            _la = self._input.LA(1)
            if _la==ClingoParser.T__8:
                self.state = 62
                self.optimization()


            self.state = 65
            self.calls()
            self.state = 66
            self.time()
            self.state = 67
            self.cpuTime()
        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx

    class ModelsContext(ParserRuleContext):

        def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
            super().__init__(parent, invokingState)
            self.parser = parser

        def TEXT(self):
            return self.getToken(ClingoParser.TEXT, 0)

        def getRuleIndex(self):
            return ClingoParser.RULE_models

        def enterRule(self, listener:ParseTreeListener):
            if hasattr( listener, "enterModels" ):
                listener.enterModels(self)

        def exitRule(self, listener:ParseTreeListener):
            if hasattr( listener, "exitModels" ):
                listener.exitModels(self)




    def models(self):

        localctx = ClingoParser.ModelsContext(self, self._ctx, self.state)
        self.enterRule(localctx, 10, self.RULE_models)
        self._la = 0 # Token type
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 69
            self.match(ClingoParser.T__4)
            self.state = 70
            self.match(ClingoParser.T__5)
            self.state = 71
            self.match(ClingoParser.TEXT)
            self.state = 73
            self._errHandler.sync(self)
            _la = self._input.LA(1)
            if _la==ClingoParser.T__6:
                self.state = 72
                self.match(ClingoParser.T__6)


        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx

    class OptimumContext(ParserRuleContext):

        def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
            super().__init__(parent, invokingState)
            self.parser = parser

        def TEXT(self):
            return self.getToken(ClingoParser.TEXT, 0)

        def getRuleIndex(self):
            return ClingoParser.RULE_optimum

        def enterRule(self, listener:ParseTreeListener):
            if hasattr( listener, "enterOptimum" ):
                listener.enterOptimum(self)

        def exitRule(self, listener:ParseTreeListener):
            if hasattr( listener, "exitOptimum" ):
                listener.exitOptimum(self)




    def optimum(self):

        localctx = ClingoParser.OptimumContext(self, self._ctx, self.state)
        self.enterRule(localctx, 12, self.RULE_optimum)
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 75
            self.match(ClingoParser.T__7)
            self.state = 76
            self.match(ClingoParser.T__5)
            self.state = 77
            self.match(ClingoParser.TEXT)
        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx

    class OptimizationContext(ParserRuleContext):

        def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
            super().__init__(parent, invokingState)
            self.parser = parser

        def TEXT(self):
            return self.getToken(ClingoParser.TEXT, 0)

        def getRuleIndex(self):
            return ClingoParser.RULE_optimization

        def enterRule(self, listener:ParseTreeListener):
            if hasattr( listener, "enterOptimization" ):
                listener.enterOptimization(self)

        def exitRule(self, listener:ParseTreeListener):
            if hasattr( listener, "exitOptimization" ):
                listener.exitOptimization(self)




    def optimization(self):

        localctx = ClingoParser.OptimizationContext(self, self._ctx, self.state)
        self.enterRule(localctx, 14, self.RULE_optimization)
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 79
            self.match(ClingoParser.T__8)
            self.state = 80
            self.match(ClingoParser.T__5)
            self.state = 81
            self.match(ClingoParser.TEXT)
        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx

    class CallsContext(ParserRuleContext):

        def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
            super().__init__(parent, invokingState)
            self.parser = parser

        def TEXT(self):
            return self.getToken(ClingoParser.TEXT, 0)

        def getRuleIndex(self):
            return ClingoParser.RULE_calls

        def enterRule(self, listener:ParseTreeListener):
            if hasattr( listener, "enterCalls" ):
                listener.enterCalls(self)

        def exitRule(self, listener:ParseTreeListener):
            if hasattr( listener, "exitCalls" ):
                listener.exitCalls(self)




    def calls(self):

        localctx = ClingoParser.CallsContext(self, self._ctx, self.state)
        self.enterRule(localctx, 16, self.RULE_calls)
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 83
            self.match(ClingoParser.T__9)
            self.state = 84
            self.match(ClingoParser.T__5)
            self.state = 85
            self.match(ClingoParser.TEXT)
        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx

    class TimeContext(ParserRuleContext):

        def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
            super().__init__(parent, invokingState)
            self.parser = parser

        def TEXT(self, i:int=None):
            if i is None:
                return self.getTokens(ClingoParser.TEXT)
            else:
                return self.getToken(ClingoParser.TEXT, i)

        def getRuleIndex(self):
            return ClingoParser.RULE_time

        def enterRule(self, listener:ParseTreeListener):
            if hasattr( listener, "enterTime" ):
                listener.enterTime(self)

        def exitRule(self, listener:ParseTreeListener):
            if hasattr( listener, "exitTime" ):
                listener.exitTime(self)




    def time(self):

        localctx = ClingoParser.TimeContext(self, self._ctx, self.state)
        self.enterRule(localctx, 18, self.RULE_time)
        self._la = 0 # Token type
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 88 
            self._errHandler.sync(self)
            _alt = 1
            while _alt!=2 and _alt!=ATN.INVALID_ALT_NUMBER:
                if _alt == 1:
                    self.state = 87
                    _la = self._input.LA(1)
                    if not((((_la) & ~0x3f) == 0 and ((1 << _la) & ((1 << ClingoParser.T__2) | (1 << ClingoParser.T__3) | (1 << ClingoParser.T__5) | (1 << ClingoParser.TEXT))) != 0)):
                        self._errHandler.recoverInline(self)
                    else:
                        self._errHandler.reportMatch(self)
                        self.consume()

                else:
                    raise NoViableAltException(self)
                self.state = 90 
                self._errHandler.sync(self)
                _alt = self._interp.adaptivePredict(self._input,9,self._ctx)

        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx

    class CpuTimeContext(ParserRuleContext):

        def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
            super().__init__(parent, invokingState)
            self.parser = parser

        def TEXT(self, i:int=None):
            if i is None:
                return self.getTokens(ClingoParser.TEXT)
            else:
                return self.getToken(ClingoParser.TEXT, i)

        def getRuleIndex(self):
            return ClingoParser.RULE_cpuTime

        def enterRule(self, listener:ParseTreeListener):
            if hasattr( listener, "enterCpuTime" ):
                listener.enterCpuTime(self)

        def exitRule(self, listener:ParseTreeListener):
            if hasattr( listener, "exitCpuTime" ):
                listener.exitCpuTime(self)




    def cpuTime(self):

        localctx = ClingoParser.CpuTimeContext(self, self._ctx, self.state)
        self.enterRule(localctx, 20, self.RULE_cpuTime)
        self._la = 0 # Token type
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 93 
            self._errHandler.sync(self)
            _la = self._input.LA(1)
            while True:
                self.state = 92
                _la = self._input.LA(1)
                if not((((_la) & ~0x3f) == 0 and ((1 << _la) & ((1 << ClingoParser.T__2) | (1 << ClingoParser.T__3) | (1 << ClingoParser.T__5) | (1 << ClingoParser.TEXT))) != 0)):
                    self._errHandler.recoverInline(self)
                else:
                    self._errHandler.reportMatch(self)
                    self.consume()
                self.state = 95 
                self._errHandler.sync(self)
                _la = self._input.LA(1)
                if not ((((_la) & ~0x3f) == 0 and ((1 << _la) & ((1 << ClingoParser.T__2) | (1 << ClingoParser.T__3) | (1 << ClingoParser.T__5) | (1 << ClingoParser.TEXT))) != 0)):
                    break

        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx








import pandas as pd
import numpy as np
import sqlite3
from .pwe_helper import lineno, isfloat, mkdir_p, PossibleWorld, Relation


def rel_id_from_rel_name(rel_name, relations):
    """
    Arguments:
    1. rel_name: name of the relation
    2. relations: list of Relation objects

    Returns:
    Relation ID (r_id) of the Relation object with the given relation name.

    Description:
    Given a list of Relation objects and a relation name, returns the relation id of
    the Relation object with the given relation name.
    """
    for i, rel in enumerate(relations):
        if rel.relation_name == rel_name:
            if i != rel.r_id:
                print("Relations not in order")
            return rel.r_id

    return None


def union_panda(dfs, pws, relations, rl_id=0, col_names=[], pws_to_consider=[], do_print=True):
    """
    Arguments:
    1. dfs: List of Panda DataFrame Objects
    2. pws: List of the PossibleWorld Objects
    3. relations: List of Relation Objects
    4. rl_id: (Optional, default = 0) The Relation ID of the relation over which the union is to be done.
    5. col_names: (Optional, defaults to all the columns) A list columns in the given relation over which to
    perform the union operation.
    6. pws_to_consider: (Optional, defualts to all the PWs) A list of PW IDs of the Possible Worlds over which
    to perform the union operation.
    7. do_print: (Optional, defaults to True) Boolean flag whether or not to print the result of the union.

    Returns:
    A Panda DataFrame with the Union of the given PWs on the provided relation on the given list of columns.
    """
    expected_pws = len(pws)

    if not pws_to_consider:
        pws_to_consider = [j for j in range(1, expected_pws + 1)]

    if not col_names:
        col_names = list(dfs[rl_id])[1:]

    df = dfs[rl_id]
    s1 = df[df.pw == pws_to_consider[0]]
    for j in pws_to_consider[1:]:
        s1 = pd.merge(s1, df[df.pw == j], how='outer', on=col_names)
    s1 = s1[col_names]

    if do_print:
        print("Union for the relation", str(relations[rl_id].relation_name), "on features",
              str(', '.join(map(str, col_names))), "for PWs", str(', '.join(map(str, pws_to_consider))))
        if len(s1) > 0:
            print(str(s1))
        else:
            print("NULL")

    return s1


# 1: Intersection
def intersection_sqlite(dfs, pws, relations, conn, rl_id=0, col_names=[], pws_to_consider=[], do_print=True):
    expected_pws = len(pws)

    if pws_to_consider == []:
        pws_to_consider = [j for j in range(1, expected_pws + 1)]

    if col_names == []:
        col_names = list(dfs[rl_id])[1:]
    query_intersection = ''

    col_names = ', '.join(map(str, col_names))
    for j in pws_to_consider[:-1]:
        query_intersection += 'select ' + col_names + ' from ' + str(
            relations[rl_id].relation_name) + ' where pw = ' + str(j) + ' intersect '
    query_intersection += 'select ' + col_names + ' from ' + str(relations[rl_id].relation_name) + ' where pw = ' + str(
        pws_to_consider[-1]) + ';'
    ik = pd.read_sql_query(query_intersection, conn)
    if do_print:
        print("Intersection for the relation " + str(
            relations[rl_id].relation_name) + " on features " + col_names + " for PWs " + str(
            ', '.join(map(str, pws_to_consider))))
        if len(ik) > 0:
            print(str(ik))
        else:
            print("NULL")

    return ik


# 2: Union
def union_sqlite(dfs, pws, relations, conn, rl_id=0, col_names=[], pws_to_consider=[], do_print=True):
    expected_pws = len(pws)

    if not pws_to_consider:
        pws_to_consider = [j for j in range(1, expected_pws + 1)]

    if not col_names:
        col_names = list(dfs[rl_id])[1:]
    query_union = ''

    col_names = ', '.join(map(str, col_names))
    for j in pws_to_consider[:-1]:
        query_union += 'select ' + col_names + ' from ' + str(relations[rl_id].relation_name) + ' where pw = ' + str(
            j) + ' union '
    query_union += 'select ' + col_names + ' from ' + str(relations[rl_id].relation_name) + ' where pw = ' + str(
        pws_to_consider[-1]) + ';'
    ik = pd.read_sql_query(query_union, conn)
    if do_print:
        print("Union for the relation ", str(relations[rl_id].relation_name), " on features ", col_names, " for PWs ",
              str(', '.join(map(str, pws_to_consider))))
        if len(ik) > 0:
            print(str(ik))
        else:
            print("NULL")

    return ik


# 3: Frequency of a tuple
def freq_sqlite(dfs, pws, relations, conn, rl_id=0, col_names=[], values=[], pws_to_consider=[], do_print=True):
    expected_pws = len(pws)
    all_tuples = None
    freqs = []

    if not pws_to_consider:
        pws_to_consider = [j for j in range(1, expected_pws + 1)]

    if col_names != [] and values != [] and len(col_names) != len(values):
        print('Lengths of col_names and values don\'t match.')
        return None

    if not col_names:
        col_names = list(dfs[rl_id])[1:]

    if not values:
        all_tuples = union_panda(dfs, pws, relations, rl_id, col_names, pws_to_consider, False)
    else:
        k = [values]
        all_tuples = pd.DataFrame(k, columns=col_names)

    for j in range(len(all_tuples)):
        query = 'select count(*) from ' + str(relations[rl_id].relation_name) + ' where '
        for k in range(len(col_names)):
            query += col_names[k] + '=' + "'" + all_tuples.ix[j][k] + "'" + ' and '

        query += 'pw in (' + str(', '.join(map(str, pws_to_consider))) + ');'
        # print query
        ik = pd.read_sql_query(query, conn)
        if do_print:
            print("Frequency of tuple " + str(tuple(all_tuples.ix[j])) + ' of the relation ' + str(
                relations[rl_id].relation_name) + ' for attributes ' + str(
                ', '.join(map(str, col_names))) + ' in PWs ' + str(
                ', '.join(map(str, pws_to_consider))) + " is: " + str(ik.ix[0][0]))
        freqs.append(ik.ix[0][0])

    return all_tuples, freqs


# 4: Number of tuples of a relation in a PW
def num_tuples_sqlite(relations, conn, rl_id, pw_id, do_print=True):
    query = 'select count(*) from ' + str(relations[rl_id].relation_name) + ' where pw = ' + str(pw_id) + ';'
    c = pd.read_sql_query(query, conn)

    if do_print:
        print("There exist " + str(c.ix[0][0]) + " tuples of relation " + str(
            relations[rl_id].relation_name) + " in PW " + str(pw_id))
    return c.ix[0][0]


# 5: Difference Query
def difference_sqlite(dfs, relations, conn, rl_id, pw_id_1, pw_id_2, col_names=[], do_print=True):

    if not col_names:
        col_names = list(dfs[rl_id])[1:]

    col_names = str(', '.join(map(str, col_names)))

    query = 'select * from (select ' + col_names + ' from ' + str(
        relations[rl_id].relation_name) + ' where pw =  ' + str(
        pw_id_1) + ') except select * from (select ' + col_names + ' from ' + str(
        relations[rl_id].relation_name) + ' where pw = ' + str(pw_id_2) + ');'
    diff = pd.read_sql_query(query, conn)

    if do_print:
        print(
            "Following is the difference between PWs {} and {} in features {} "
            "of relation {}\n".format(pw_id_1, pw_id_2, col_names, str(relations[rl_id].relation_name)))
        print(str(diff))
    return diff


def difference_both_ways_sqlite(dfs, relations, conn, rl_id, pw_id_1, pw_id_2, col_names=[], do_print=True):

    if not col_names:
        col_names = list(dfs[rl_id])[1:]

    x1 = difference_sqlite(dfs, relations, conn, rl_id, pw_id_1, pw_id_2, col_names, False)
    x2 = difference_sqlite(dfs, relations, conn, rl_id, pw_id_2, pw_id_1, col_names, False)

    diff = x1.append(x2, ignore_index=True)

    if do_print:
        print("Following tuples are in one of PW {} or {}, but not both, for relation {} and features {}\n".format(
            pw_id_1, pw_id_2, str(relations[rl_id].relation_name), str(', '.join(map(str, col_names)))))
        print(str(diff))
    return diff


# 6: Redundant Column Query
def redundant_column_sqlite(dfs, pws, relations, conn, rl_id=0, col_names=[], pws_to_consider=[], do_print=True):
    expected_pws = len(pws)

    if pws_to_consider == []:
        pws_to_consider = [j for j in range(1, expected_pws + 1)]

    if col_names == []:
        col_names = list(dfs[rl_id])[1:]

    # PW specific:
    redundant_pw_specific = []
    for i in pws_to_consider:
        for ft in col_names:
            query = 'select count(distinct ' + str(ft) + ') from (select * from ' + str(
                relations[rl_id].relation_name) + ' where pw = ' + str(i) + ');'
            k = pd.read_sql_query(query, conn)
            if int(k.ix[0][0]) <= 1:
                redundant_pw_specific.append((i, rl_id, ft))
                if do_print:
                    print('Column' + str(ft) + 'is redundant in relation' + str(
                        relations[rl_id].relation_name) + 'in PW' + str(i))

    # Across all PWs:
    redundant_across_pws = []
    for ft in col_names:
        query = 'select count(distinct ' + str(ft) + ') from (select * from ' + str(
            relations[rl_id].relation_name) + ' where pw in (' + str(', '.join(map(str, pws_to_consider))) + '));'
        k = pd.read_sql_query(query, conn)
        if int(k.ix[0][0]) <= 1:
            redundant_across_pws.append((pws_to_consider, rl_id, ft))
            if do_print:
                print('Column ' + str(ft) + ' is redundant in relation ' + str(
                    relations[rl_id].relation_name) + ' for PWs ' + str(', '.join(map(str, pws_to_consider))))

    return redundant_pw_specific, redundant_across_pws


# 7: Tuples occuring in exactly one PW:
def unique_tuples_sqlite(dfs, pws, relations, conn, rl_id=0, col_names=[], pws_to_consider=[], do_print=True):
    """
    Used to ouput the unique tuples in a given solution set. These are the unique tuples (attributes) in the
    selected possible worlds and the selected columns in the relation.
    """

    expected_pws = len(pws)

    if pws_to_consider == []:
        pws_to_consider = [j for j in range(1, expected_pws + 1)]

    if col_names == []:
        col_names = list(dfs[rl_id])[1:]

    relevant_tuples, freqs = freq_sqlite(dfs, pws, relations, conn, rl_id, col_names, [], pws_to_consider, False)
    unique_tuples = []

    for i, f in enumerate(freqs):
        if f == 1:
            query = 'select pw from ' + str(relations[rl_id].relation_name) + ' where '
            for k in range(len(col_names)):
                query += col_names[k] + '=' + "'" + relevant_tuples.ix[i][k] + "'" + ' and '

            query += 'pw in (' + str(', '.join(map(str, pws_to_consider))) + ');'

            unique_pw = pd.read_sql_query(query, conn)
            unique_pw = unique_pw.ix[0][0]

            unique_tuples.append((relevant_tuples.ix[i], unique_pw))

            if do_print:
                print('The unique tuple ' + str(tuple(relevant_tuples.ix[i])) + ' occurs only in PW ' + str(unique_pw))

    return unique_tuples



#!/usr/bin/env python3

from sklearn.cluster import DBSCAN
from PW_explorer.pwe_helper import rel_id_from_rel_name, mkdir_p
import matplotlib.pyplot as plt
from collections import defaultdict
import os


def get_colors_list(df, id):
    idx = list(map(int, list(df[df.pw == id].x1)))
    clrs = list(df[df.pw == id].x2)

    ordered_clrs = [x for _, x in sorted(zip(idx, clrs))]
    return ordered_clrs


def visualize(**kwargs):

    kwargs = defaultdict(lambda: None, kwargs)
    dfs = kwargs['dfs']
    relations = kwargs['relations']
    save_to_folder = kwargs['save_to_folder']
    dist_matrix = kwargs['dist_matrix']

    if relations is None or dfs is None:
        print("No objects passed.")
        exit(1)

    db = DBSCAN(metric='precomputed', eps=0.5, min_samples=1)
    labels = db.fit_predict(dist_matrix)

    unique_indices = {}
    label_counts = {}
    for i, label in enumerate(labels):
        if label not in unique_indices:
            unique_indices[label] = i
            label_counts[label] = 1
        else:
            label_counts[label] += 1

    df = dfs[rel_id_from_rel_name('col_2', relations)]

    figs = []
    for label, idx in list(unique_indices.items()):
        fig, ax = plt.subplots()
        cols = get_colors_list(df, idx + 1)
        # print(cols)
        group_size = [1 for _ in cols]
        ax.pie(group_size, colors=cols)
        my_circle = plt.Circle((0, 0), 0.7, color='white')
        p = plt.gcf()
        p.gca().add_artist(my_circle)
        ax.annotate(str(label_counts[label]), xy=(0, 0), ha='center', va='center', fontsize=35)
        figs.append(fig)

    if save_to_folder is not None:
        output_folder = os.path.join(save_to_folder, 'four_colorability_visualization')
        mkdir_p(output_folder)
        for i, fig in enumerate(figs):
            fig.savefig(os.path.join(output_folder, 'pattern_{}.png'.format(str(i))))

    return figs



from antlr4 import *
from .Antlr_Files.ClingoLexer import ClingoLexer
from .Antlr_Files.ClingoParser import ClingoParser
from .Antlr_Files.ClingoListener import ClingoListener
from .pwe_helper import isfloat, PossibleWorld, Relation
import pandas as pd
import numpy as np
from antlr4.tree.Trees import Trees


def rearrangePWSandRLS(relations, pws):
    """
    Sort the possible worlds and relations by their ids
    :return: None
    """
    relations.sort(key=lambda x: x.r_id)
    pws.sort(key=lambda x: x.pw_id)


def loadIntoPandas(relations, pws, dfs):
    """
    Populate the Pandas DF, one for each relation
    :return: None
    """
    # print lineno()
    for n, rl in enumerate(relations):
        cls = ['pw']
        cls.extend([str('x' + str(i)) for i in range(1, rl.arity + 1)])

        rws = []  # could convert into numpy if sure it's all float/int
        for m, pw in enumerate(pws):
            # print rl.r_id
            if rl.r_id < len(pw.rls):
                rl_data_pw = []
                for rl_data in pw.rls[rl.r_id]:
                    rl_data_pw.append(rl_data.copy())
                    rl_data_pw[-1].insert(0, pw.pw_id)
                rws.extend(rl_data_pw)

        df = pd.DataFrame(rws, columns=cls)
        dfs.append(df)


######################################################################################

######################################################################################

class AntlrClingoListener(ClingoListener):

    def __init__(self):
        self.pws = []
        self.relations = []
        self.expected_pws = 0
        self.curr_pw = None
        self.curr_pw_id = 1
        self.curr_rl = None
        self.curr_rl_data = None
        self.n_rls = 0
        self.dfs = []
        self.out_file = None

    def enterClingoOutput(self, ctx):
        if ctx.OPTIMUM_FOUND() is not None:
            if ctx.OPTIMUM_FOUND().getText() == 'UNSATISFIABLE':
                print("The problem is unsatisfiable")

    def enterSolution(self, ctx):
        self.curr_pw = PossibleWorld(self.n_rls, self.curr_pw_id)
        # assert curr_pw.pw_id == int(ctx.TEXT(0).getText())
        if ctx.TEXT(1) is not None:
            self.curr_pw.pw_soln = float(ctx.TEXT(1).getText()) if isfloat(ctx.TEXT(1).getText()) else ctx.TEXT(1).getText()

    def enterActual_soln(self, ctx):
        i = 0
        while ctx.TEXT(i) is not None:
            i += 1
        self.curr_rl = Relation(ctx.TEXT(i - 1).getText())

    def enterCustom_representation_soln(self, ctx):

        sol = ctx.TEXT().getText();
        self.curr_rl_data = sol.split(',')
        self.curr_rl.arity = len(self.curr_rl_data)
        rl_name_mod = str(self.curr_rl.relation_name + '_' + str(self.curr_rl.arity))
        self.curr_rl.relation_name = rl_name_mod

    def exitCustom_representation_soln(self, ctx):

        foundMatch = False
        for rl in self.relations:
            if self.curr_rl.relation_name == rl.relation_name and self.curr_rl.arity == rl.arity:
                self.curr_rl.r_id = rl.r_id
                # print rl.r_id, lineno() ##for debugging purposes
                foundMatch = True
                break

        if not foundMatch:
            newRl = Relation(self.curr_rl.relation_name)
            newRl.arity = self.curr_rl.arity
            newRl.r_id = self.n_rls
            # print n_rls, lineno()
            self.n_rls += 1
            self.relations.append(newRl)
            self.curr_rl.r_id = newRl.r_id

        self.curr_pw.add_relation(self.curr_rl.r_id, self.curr_rl_data)
        self.curr_rl = None  # could introduce bugs if passed by pointer in the upper statement, so be careful, use copy() if needed
        self.curr_rl_data = None

    def exitActual_soln(self, ctx):
        # print lineno()
        self.curr_rl = None
        self.curr_rl_data = None

    def exitSolution(self, ctx):
        # print lineno()
        self.pws.append(self.curr_pw)  # again be wary, else use .copy()
        self.curr_pw = None
        self.curr_pw_id += 1

    def enterOptimum(self, ctx):

        # print lineno()
        optimum_found = ctx.TEXT().getText()
        if optimum_found == 'yes':
            print('Optimum Solution was found')
        elif optimum_found == 'no':
            print('Optimum Solution was not found')
        else:
            print('Unexpected Output:', optimum_found)

    def enterOptimization(self, ctx):
        # print lineno()
        opt_soln = ctx.TEXT().getText()
        print('Optimized Solution is', opt_soln)

    def enterModels(self, ctx):

        # print lineno()
        num_models = ctx.TEXT().getText()
        num_models = int(num_models)
        print("Number of Models:", num_models)
        self.expected_pws = num_models

    def exitClingoOutput(self, ctx):

        # loading into pandas DF
        rearrangePWSandRLS(self.relations, self.pws)
        loadIntoPandas(self.relations, self.pws, self.dfs)


######################################################################################

def parse_clingo_output(fname):

    input_ = FileStream(fname)
    lexer = ClingoLexer(input_)

    # use this line to take input from the cmd line
    # lexer = ClingoLexer(StdinStream())

    stream = CommonTokenStream(lexer)
    parser = ClingoParser(stream)
    tree = parser.clingoOutput()
    # Use (uncomment) the line below to see the parse tree of the given input
    # print Trees.toStringTree(tree, None, parser)
    pw_analyzer = AntlrClingoListener()
    walker = ParseTreeWalker()
    walker.walk(pw_analyzer, tree)

    return pw_analyzer.dfs, pw_analyzer.relations, pw_analyzer.pws



#!/usr/bin/env python3

import errno
import inspect
import os
import pickle
import sqlite3


###################################################################

# to help debug
def lineno():
    """Returns the current line number in our program."""
    return inspect.currentframe().f_back.f_lineno


###################################################################

###################################################################

# helper funcs

def isfloat(value):
    """returns true if a value can be typecasted as a float, else false"""
    try:
        float(value)
        return True
    except ValueError:
        return False


def mkdir_p(path):
    """Make a directory if it doesn't already exist"""
    try:
        os.makedirs(path)
    except OSError as exc:  # Python >2.5
        if exc.errno == errno.EEXIST and os.path.isdir(path):
            pass
        else:
            raise


def preprocess_clingo_output(clingo_raw_output: list):

    start = 0
    for i, line in enumerate(clingo_raw_output):
        if line.strip() == 'Solving...':
            start = i + 1
            break

    return clingo_raw_output[start:]


###################################################################

class PossibleWorld:
    """
    Class to store details and solution relating to every possible world
    """
    def __init__(self, num_relations, pw_id):
        self.rls = [[] for i in range(num_relations)]
        self.pw_id = pw_id
        self.pw_soln = 0

    def add_relation(self, relation_id, relation_data):
        if relation_id >= len(self.rls):
            self.rls.append([])
        self.rls[relation_id].append(relation_data)


###################################################################

class Relation:
    """
    Class to store description of each relation found in the clingo output
    """

    def __init__(self, relation_name):
        self.relation_name = relation_name
        self.arity = 0
        self.r_id = 0


###################################################################

# File Location Constants:


SRC_TEMP_PICKLE_LOC = '.temp_pickle_data/'
CURRENT_PROJECT_NAME_LOC = SRC_TEMP_PICKLE_LOC + 'current_project_name.pkl'
PROJECT_RESULTS_LOC = 'PWE_Results/'
PROJECT_ASP_INPUT_FOLDER = 'ASP_Input/'
PROJECT_ASP_OUTPUT_FOLDER = 'ASP_Output/'
PROJECT_EXPORTS_FOLDER = 'Exports/'
PROJECT_VISUALIZATIONS_FOLDER = 'Visualizations/'
PROJECT_TEMP_PICKLE_DATA_FOLDER = 'temp_pickle_data/'
CUSTOM_DISTANCE_FUNCTIONS_FOLDER = 'Custom_Distance_Functions'
CUSTOM_VISUALIZATION_FUNCTIONS_FOLDER = 'Custom_Visualization_Functions'


def get_asp_input_folder(project_name):
    """
    :type project_name: string
    """
    loc = PROJECT_RESULTS_LOC + project_name + '/' + PROJECT_ASP_INPUT_FOLDER
    mkdir_p(loc)
    return os.path.abspath(loc)


def get_asp_output_folder(project_name):
    """
    :type project_name: string
    """
    loc = PROJECT_RESULTS_LOC + project_name + '/' + PROJECT_ASP_OUTPUT_FOLDER
    mkdir_p(loc)
    return os.path.abspath(loc)


def get_current_project_name():
    project_name = None
    try:
        with open(CURRENT_PROJECT_NAME_LOC, 'rb') as f:
            project_name = pickle.load(f)
    except IOError:
        project_name = None
    return project_name


def set_current_project_name(project_name):
    mkdir_p(SRC_TEMP_PICKLE_LOC)
    with open(CURRENT_PROJECT_NAME_LOC, 'wb') as f:
        pickle.dump(project_name, f)


def get_save_folder(project_name, data_type):
    data_types_dict = {
        'csv_export': PROJECT_EXPORTS_FOLDER+'csv/',
        'sql_export': PROJECT_EXPORTS_FOLDER+'sql/',
        'msg_export': PROJECT_EXPORTS_FOLDER + 'msg/',
        'h5_export': PROJECT_EXPORTS_FOLDER + 'h5/',
        'pkl_export': PROJECT_EXPORTS_FOLDER + 'pkl/',
        'temp_pickle_data': PROJECT_TEMP_PICKLE_DATA_FOLDER,
        'visualization': PROJECT_VISUALIZATIONS_FOLDER,
    }

    relative_loc = PROJECT_RESULTS_LOC + project_name + '/'

    if data_type in data_types_dict:
        relative_loc += data_types_dict[data_type]
    else:
        relative_loc += data_type + '/'

    mkdir_p(relative_loc)

    return os.path.abspath(relative_loc)


def get_file_save_name(project_name, file_type):

    if file_type in ['dist_matrix', 'dfs', 'relations', 'pws', 'complexities']:
        return file_type + '.pkl'
    return None


def load_from_temp_pickle(project_name, file_type):
    if file_type in ['dfs', 'pws', 'relations', 'dist_matrix', 'complexities']:

        try:
            with open(get_save_folder(project_name, 'temp_pickle_data') + '/' +
                      get_file_save_name(project_name, file_type), 'rb') as input_file:
                return pickle.load(input_file)
        except IOError:
            print("Could not find the project, check project/session name entered.")
            exit(1)


def get_sql_conn(project_name):
    try:
        conn = sqlite3.connect(get_save_folder(project_name, 'sql_export') + '/' +str(project_name) + ".db")
        return conn
    except sqlite3.Error:
        print("Could not find the associated sqlite database. Please recheck project_name " \
              "or make sure a sql db has been exported using export module")
        exit(1)


def rel_id_from_rel_name(rel_name, relations):
    for i, rel in enumerate(relations):
        if rel.relation_name == rel_name:
            if i != rel.r_id:
                print("Relations not in order")
            return rel.r_id
    return None



#!/usr/bin/env python3
# SQLite QUERY SCRIPT:

import pandas as pd
import numpy as np
import sqlite3
import os
import argparse
from .pwe_helper import get_current_project_name, set_current_project_name, \
    load_from_temp_pickle, get_sql_conn, rel_id_from_rel_name
from . import sql_funcs


def union_panda(relations, expected_pws, dfs, pws, rl_id=0, col_names=[], pws_to_consider=[], do_print=True):

    if not pws_to_consider:
        pws_to_consider = [j for j in range(1, expected_pws + 1)]

    if not col_names:
        col_names = list(dfs[rl_id])[1:]

    return sql_funcs.union_panda(dfs, pws, relations, rl_id, col_names, pws_to_consider, do_print)


# 1: Intersection
def intersection_sqlite(relations, expected_pws, dfs, conn, pws, rl_id=0, col_names=[],
                        pws_to_consider=[], do_print=True):

    if not pws_to_consider:
        pws_to_consider = [j for j in range(1, expected_pws + 1)]

    if not col_names:
        col_names = list(dfs[rl_id])[1:]
    return sql_funcs.intersection_sqlite(dfs, pws, relations, conn, rl_id, col_names, pws_to_consider, do_print)


# 2: Union
def union_sqlite(relations, expected_pws, dfs, conn, pws, rl_id=0, col_names=[], pws_to_consider=[], do_print=True):

    if not pws_to_consider:
        pws_to_consider = [j for j in range(1, expected_pws + 1)]

    if not col_names:
        col_names = list(dfs[rl_id])[1:]

    return sql_funcs.union_sqlite(dfs, pws, relations, conn, rl_id, col_names, pws_to_consider, do_print)


# 3: Frequency of a tuple
def freq_sqlite(relations, dfs, conn, pws, rl_id=0, col_names=[], values=[], pws_to_consider=[],
                do_print=True):

    return sql_funcs.freq_sqlite(dfs, pws, relations, conn, rl_id, col_names, values, pws_to_consider, do_print)


# 4: Number of tuples of a relation in a PW
def num_tuples_sqlite(relations, conn, rl_id, pw_id, do_print=True):

    return sql_funcs.num_tuples_sqlite(relations, conn, rl_id, pw_id, do_print)


# 5: Difference Query
def difference_sqlite(relations, conn, dfs, rl_id, pw_id_1, pw_id_2, col_names=[], do_print=True):

    return sql_funcs.difference_sqlite(dfs, relations, conn, rl_id, pw_id_1, pw_id_2, col_names, do_print)


def difference_both_ways_sqlite(relations, conn, dfs, rl_id, pw_id_1, pw_id_2, col_names=[], do_print=True):

    return sql_funcs.difference_both_ways_sqlite(dfs, relations, conn, rl_id, pw_id_1, pw_id_2, col_names, do_print)


# 6: Redundant Column Query
def redundant_column_sqlite(relations, expected_pws, dfs, conn, pws, rl_id=0, col_names=[], pws_to_consider=[],
                            do_print=True):

    if not pws_to_consider:
        pws_to_consider = [j for j in range(1, expected_pws + 1)]

    return sql_funcs.redundant_column_sqlite(dfs, pws, relations, conn, rl_id, col_names, pws_to_consider, do_print)


# 7: Tuples occuring in exactly one PW:
def unique_tuples_sqlite(relations, pws, dfs, conn, rl_id=0, col_names=None, pws_to_consider=None, do_print=True):

    return sql_funcs.unique_tuples_sqlite(dfs, pws, relations, conn, rl_id, col_names, pws_to_consider, do_print)


def __main__():
    parser = argparse.ArgumentParser()
    parser.add_argument("-p", "--project_name", type=str, help="provide session/project name used while parsing")
    group = parser.add_mutually_exclusive_group()
    group.add_argument("-intersection", action='store_true', default=False,
                       help="provide either relation name or relation_id using the -rel_name or -rel_id flag respectively, "
                            "columns to consider using the -cols flag and possible worlds to consider using the -pws flag.")
    group.add_argument("-union", action='store_true', default=False,
                       help="provide either relation name or relation_id using the -rel_name or -rel_id flag respectively, "
                            "columns to consider using the -cols flag and possible worlds to consider using the -pws flag.")
    group.add_argument("-freq", action='store_true', default=False,
                       help="provide either relation name or relation_id using the -rel_name or -rel_id flag respectively, "
                            "columns to consider using the -cols flag, possible worlds to consider using the -pws flag and "
                            "the values for the columns (in the mentioned order) using the -vals flag (optional).")
    group.add_argument("-num_tuples", action='store_true', default=False,
                       help="provide either relation name or relation_id using the -rel_name or -rel_id flag respectively "
                            "and the possible world ids to count the tuples in using the -pws flag.")
    group.add_argument("-difference", choices=('one-way', 'symmetric'),
                       help="provide either relation name or relation_id using the -rel_name or -rel_id flag respectively, "
                            "columns to consider using the -cols flag and the two possible world ids using the -pws flag.")
    group.add_argument("-redundant_column", action='store_true', default=False,
                       help="provide either relation name or relation_id using the -rel_name or -rel_id flag respectively, "
                            "columns to consider using the -cols flag and possible worlds to consider using the -pws flag.")
    group.add_argument("-unique_tuples", action='store_true', default=False,
                       help="provide either relation name or relation_id using the -rel_name or -rel_id flag respectively, "
                            "columns to consider using the -cols flag and possible worlds to consider using the -pws flag.")
    group.add_argument("-custom", type=str, help="provide the query enclosed in '' .")
    group.add_argument("-custom_file", type=str, help="provide the .sql file containing the query.")
    group.add_argument("-show_relations", action='store_true', default=False,
                       help="to get a list of relations and corresponding relation ids.")

    parser.add_argument("-rel_name", type=str,
                        help="provide the relation name to query. Note that if both rel_id and rel_name are provided, "
                             "rel_name is disregarded.")
    parser.add_argument("-rel_id", type=int,
                        help="provide the relation id of the relation to query. To view relation ids, use -show_relations")
    parser.add_argument("-cols", type=str, nargs='*', default=[],
                        help="provide the columns of the selected relations to consider for the chosen query. If you want "
                             "to consider all the columns, do not include this flag.")
    parser.add_argument("-pws", type=int, nargs='*', default=[],
                        help="provide the possible world ids of the possible world to consider for this query. If you want "
                             "to consider all the possible worlds, do not include this flag. Please note that difference "
                             "query requires exactly 2 arguments for this flag.")
    parser.add_argument("-vals", nargs='*', default=[], type=str,
                        help="provide the values for the freq query in the same order as the mentioned columns. If you "
                             "want to query all possible tuples, do not include this flag.")

    args = parser.parse_args()

    project_name = ''
    if args.project_name is None:
        project_name = get_current_project_name()
        if project_name is None:
            print("Couldn't find current project. Please provide a project name.")
            exit(1)
    else:
        project_name = args.project_name

    dfs = load_from_temp_pickle(project_name, 'dfs')
    relations = load_from_temp_pickle(project_name, 'relations')
    pws = load_from_temp_pickle(project_name, 'pws')
    conn = get_sql_conn(project_name)
    expected_pws = len(pws)
    # print "pws: ", args.pws
    # print "cols: ", args.cols

    if args.intersection:

        if args.rel_name is None and args.rel_id is None:
            print("Please include either the -rel_name or -rel_id flag along with the appropriate argument.")
            exit(0)

        r_id = args.rel_id
        if r_id is None:
            r_id = rel_id_from_rel_name(args.rel_name, relations)

        soln = None
        try:
            soln = intersection_sqlite(r_id, args.cols, args.pws, True)
        except Exception as e:
            print("Query failed. Please check the provided arguments to make sure they are valid.")
            print("Error: ", str(e))
            exit(1)

    elif args.union:

        if args.rel_name is None and args.rel_id is None:
            print("Please include either the -rel_name or -rel_id flag along with the appropriate argument.")
            exit(0)

        r_id = args.rel_id
        if r_id is None:
            r_id = rel_id_from_rel_name(args.rel_name, relations)

        soln = None
        try:
            soln = union_sqlite(r_id, args.cols, args.pws, True)
        except Exception as e:
            print("Query failed. Please check the provided arguments to make sure they are valid.")
            print("Error: ", str(e))
            exit(1)

    elif args.freq:

        if args.rel_name is None and args.rel_id is None:
            print("Please include either the -rel_name or -rel_id flag along with the appropriate argument.")
            exit(0)

        r_id = args.rel_id
        if r_id is None:
            r_id = rel_id_from_rel_name(args.rel_name, relations)

        soln = None
        try:
            soln = freq_sqlite(r_id, args.cols, args.vals, args.pws, True)
        except Exception as e:
            print("Query failed. Please check the provided arguments to make sure they are valid.")
            print("Error: ", str(e))
            exit(1)

    elif args.num_tuples:

        if args.rel_name is None and args.rel_id is None:
            print("Please include either the -rel_name or -rel_id flag along with the appropriate argument.")
            exit(0)

        r_id = args.rel_id
        if r_id is None:
            r_id = rel_id_from_rel_name(args.rel_name, relations)

        pws_to_consider = args.pws
        if pws_to_consider == []:
            pws_to_consider = [j for j in range(1, expected_pws + 1)]

        soln = []
        try:
            for i in pws_to_consider:
                soln.append(num_tuples_sqlite(r_id, i, True))
        except Exception as e:
            print("Query failed. Please check the provided arguments to make sure they are valid.")
            print("Error: ", str(e))
            exit(1)

    elif args.difference is not None:

        if args.rel_name is None and args.rel_id is None:
            print("Please include either the -rel_name or -rel_id flag along with the appropriate argument.")
            exit(0)

        r_id = args.rel_id
        if r_id is None:
            r_id = rel_id_from_rel_name(args.rel_name, relations)

        if len(args.pws) != 2:
            print("Please provide exactly 2 possible world ids.")
            exit(0)

        soln = None

        if args.difference == 'one-way':
            try:
                soln = difference_sqlite(r_id, args.pws[0], args.pws[1], args.cols, True)
            except Exception as e:
                print("Query failed. Please check the provided arguments to make sure they are valid.")
                print("Error: ", str(e))
                exit(1)
        elif args.difference == 'symmetric':
            try:
                soln = difference_both_ways_sqlite(r_id, args.pws[0], args.pws[1], args.cols, True)
            except Exception as e:
                print("Query failed. Please check the provided arguments to make sure they are valid.")
                print("Error: ", str(e))
                exit(1)

    elif args.redundant_column:

        if args.rel_name is None and args.rel_id is None:
            print("Please include either the -rel_name or -rel_id flag along with the appropriate argument.")
            exit(0)

        r_id = args.rel_id
        if r_id is None:
            r_id = rel_id_from_rel_name(args.rel_name, relations)

        soln = None
        try:
            soln = redundant_column_sqlite(r_id, args.cols, args.pws, True)
        except Exception as e:
            print("Query failed. Please check the provided arguments to make sure they are valid.")
            print("Error: ", str(e))
            exit(1)

    elif args.unique_tuples:

        if args.rel_name is None and args.rel_id is None:
            print("Please include either the -rel_name or -rel_id flag along with the appropriate argument.")
            exit(0)

        r_id = args.rel_id
        if r_id is None:
            r_id = rel_id_from_rel_name(args.rel_name, relations)

        soln = None
        try:
            soln = unique_tuples_sqlite(r_id, args.cols, args.pws, True)
        except Exception as e:
            print("Query failed. Please check the provided arguments to make sure they are valid.")
            print("Error: ", str(e))
            exit(1)

    elif args.show_relations:

        print('Following are the parsed relation IDs and relation names:')
        for i, rl in enumerate(relations):
            print(str(i) + ':', str(rl.relation_name))

    elif args.custom is not None:

        ik = None
        try:
            ik = pd.read_sql_query(args.custom, conn)
            print(str(ik))
            if len(ik) <= 0:
                print("NULL")
        except Exception as e:
            print("Query failed. Please check the provided query to make sure it is valid.")
            print("Error: ", str(e))
            exit(1)

    elif args.custom_file is not None:

        if not os.path.exists(args.custom_file):
            print("No file by the name {} exists. Please recheck the .sql file location".format(args.custom_file))
            exit(1)

        f = open(args.custom_file, 'r')
        f = f.read()

        sqlQueries = f.split(';')
        soln = []
        # print sqlQueries
        for q in sqlQueries:
            if q.strip() == '':
                continue
            try:
                soln.append(pd.read_sql_query(q, conn))
                print(str(soln[-1]))
                if len(soln[-1]) <= 0:
                    print("NULL")
            except Exception as e:
                print("Query failed. Please check the provided query to make sure it is valid.")
                print("Error: ", str(e))
                exit(1)

    conn.commit()
    conn.close()
    set_current_project_name(project_name)


if __name__ == '__main__':
    __main__()


# Generated from Clingo.g4 by ANTLR 4.7.1
from antlr4 import *
if __name__ is not None and "." in __name__:
    from .ClingoParser import ClingoParser
else:
    from ClingoParser import ClingoParser

# This class defines a complete listener for a parse tree produced by ClingoParser.
class ClingoListener(ParseTreeListener):

    # Enter a parse tree produced by ClingoParser#clingoOutput.
    def enterClingoOutput(self, ctx:ClingoParser.ClingoOutputContext):
        pass

    # Exit a parse tree produced by ClingoParser#clingoOutput.
    def exitClingoOutput(self, ctx:ClingoParser.ClingoOutputContext):
        pass


    # Enter a parse tree produced by ClingoParser#solution.
    def enterSolution(self, ctx:ClingoParser.SolutionContext):
        pass

    # Exit a parse tree produced by ClingoParser#solution.
    def exitSolution(self, ctx:ClingoParser.SolutionContext):
        pass


    # Enter a parse tree produced by ClingoParser#actual_soln.
    def enterActual_soln(self, ctx:ClingoParser.Actual_solnContext):
        pass

    # Exit a parse tree produced by ClingoParser#actual_soln.
    def exitActual_soln(self, ctx:ClingoParser.Actual_solnContext):
        pass


    # Enter a parse tree produced by ClingoParser#custom_representation_soln.
    def enterCustom_representation_soln(self, ctx:ClingoParser.Custom_representation_solnContext):
        pass

    # Exit a parse tree produced by ClingoParser#custom_representation_soln.
    def exitCustom_representation_soln(self, ctx:ClingoParser.Custom_representation_solnContext):
        pass


    # Enter a parse tree produced by ClingoParser#summary.
    def enterSummary(self, ctx:ClingoParser.SummaryContext):
        pass

    # Exit a parse tree produced by ClingoParser#summary.
    def exitSummary(self, ctx:ClingoParser.SummaryContext):
        pass


    # Enter a parse tree produced by ClingoParser#models.
    def enterModels(self, ctx:ClingoParser.ModelsContext):
        pass

    # Exit a parse tree produced by ClingoParser#models.
    def exitModels(self, ctx:ClingoParser.ModelsContext):
        pass


    # Enter a parse tree produced by ClingoParser#optimum.
    def enterOptimum(self, ctx:ClingoParser.OptimumContext):
        pass

    # Exit a parse tree produced by ClingoParser#optimum.
    def exitOptimum(self, ctx:ClingoParser.OptimumContext):
        pass


    # Enter a parse tree produced by ClingoParser#optimization.
    def enterOptimization(self, ctx:ClingoParser.OptimizationContext):
        pass

    # Exit a parse tree produced by ClingoParser#optimization.
    def exitOptimization(self, ctx:ClingoParser.OptimizationContext):
        pass


    # Enter a parse tree produced by ClingoParser#calls.
    def enterCalls(self, ctx:ClingoParser.CallsContext):
        pass

    # Exit a parse tree produced by ClingoParser#calls.
    def exitCalls(self, ctx:ClingoParser.CallsContext):
        pass


    # Enter a parse tree produced by ClingoParser#time.
    def enterTime(self, ctx:ClingoParser.TimeContext):
        pass

    # Exit a parse tree produced by ClingoParser#time.
    def exitTime(self, ctx:ClingoParser.TimeContext):
        pass


    # Enter a parse tree produced by ClingoParser#cpuTime.
    def enterCpuTime(self, ctx:ClingoParser.CpuTimeContext):
        pass

    # Exit a parse tree produced by ClingoParser#cpuTime.
    def exitCpuTime(self, ctx:ClingoParser.CpuTimeContext):
        pass





from . import run_clingo
from . import load_worlds
from . import pwe_query
from . import complexity_calc
from . import dist_calc
from . import export
from . import visualize


__all__ = ['run_clingo', 'load_worlds', 'pwe_query', 'complexity_calc', 'dist_calc', 'export', 'visualize']


#!/usr/bin/env python3

import pandas as pd


class PWEQuery:

    @staticmethod
    def intersection(relations: list, expected_pws: int, dfs: list, rl_id=0, col_names: list=None,
                     pws_to_consider: list=None, do_print: bool=True):

        if not pws_to_consider:
            pws_to_consider = [j for j in range(1, expected_pws + 1)]

        if not col_names:
            col_names = list(dfs[rl_id])[1:]

        df = dfs[rl_id]
        s1 = df[df.pw == pws_to_consider[0]]
        for j in pws_to_consider[1:]:
            s1 = pd.merge(s1, df[df.pw == j], how='inner', on=col_names)
        s1 = s1[col_names]
        if do_print:
            print("Intersection for the relation", str(relations[rl_id].relation_name), "on features",
                  str(', '.join(map(str, col_names))), "for PWs", str(', '.join(map(str, pws_to_consider))))
            if len(s1) > 0:
                print(str(s1))
            else:
                print("NULL")

        return s1

    @staticmethod
    def union(relations: list, expected_pws: int, dfs: list, rl_id: int=0, col_names: list=None,
              pws_to_consider: list=None, do_print: bool=True):

        if not pws_to_consider:
            pws_to_consider = [j for j in range(1, expected_pws + 1)]

        if not col_names:
            col_names = list(dfs[rl_id])[1:]

        df = dfs[rl_id]
        s1 = df[df.pw == pws_to_consider[0]]
        for j in pws_to_consider[1:]:
            s1 = pd.merge(s1, df[df.pw == j], how='outer', on=col_names)
        s1 = s1[col_names]

        if do_print:
            print("Intersection for the relation", str(relations[rl_id].relation_name), "on features",
                  str(', '.join(map(str, col_names))), "for PWs", str(', '.join(map(str, pws_to_consider))))
            if len(s1) > 0:
                print(str(s1))
            else:
                print("NULL")

        return s1

    @staticmethod
    def freq(relations: list, expected_pws: int, dfs: list, rl_id: int=0, col_names: list=None, values: list=None,
                   pws_to_consider: list=None, do_print: bool=True):

        all_tuples = None
        freqs = []

        if not pws_to_consider:
            pws_to_consider = [j for j in range(1, expected_pws + 1)]

        if col_names and values and len(col_names) != len(values):
            print('Lengths of col_names and values don\'t match.')
            return None

        if not col_names:
            col_names = list(dfs[rl_id])[1:]

        if not values:
            all_tuples = PWEQuery.union(relations, expected_pws, dfs, rl_id, col_names, pws_to_consider, False)
        else:
            k = [values]
            all_tuples = pd.DataFrame(k, columns=col_names)

        df = dfs[rl_id]
        for j in range(len(all_tuples)):
            expr = ''
            for k in range(len(col_names) - 1):
                expr += str(col_names[k]) + ' == ' + "'" + str(all_tuples.ix[j][k]) + "'" + ' and '
            expr += str(col_names[-1]) + ' == ' + "'" + str(all_tuples.ix[j][-1]) + "'"
            expr += ' and pw in [' + str(', '.join(map(str, pws_to_consider))) + ']'
            s3 = df.query(expr)
            tmp = len(s3)
            if do_print:
                print("Frequency of tuple " + str(tuple(all_tuples.ix[j])) + ' of the relation ' + str(
                    relations[rl_id].relation_name) + ' for attributes ' + str(
                    ', '.join(map(str, col_names))) + ' in PWs ' + str(
                    ', '.join(map(str, pws_to_consider))) + " is: " + str(tmp))
            freqs.append(tmp)

        return all_tuples, freqs

    @staticmethod
    def num_tuples(relations: list, dfs: list, rl_id: int, pw_id: int, do_print: bool=True):

        df = dfs[rl_id]
        c = len(df[df.pw == pw_id])
        if do_print:
            print("There exist " + str(c) + " tuples of relation " + str(relations[rl_id].relation_name) + " in PW " + str(
                pw_id))
        return c

    @staticmethod
    def difference(relations: list, dfs: list, rl_id: int, pw_id_1: int, pw_id_2: int, col_names: list=None,
                         do_print: bool=True):

        if not col_names:
            col_names = list(dfs[rl_id])[1:]

        df = dfs[rl_id]
        x1 = df[df.pw == pw_id_1][col_names]
        x2 = df[df.pw == pw_id_2][col_names]

        diff = pd.concat([x1, x2, x2]).drop_duplicates(keep=False)
        if do_print:
            print(
                "Following is the difference between PWs {} and {} in features {} of "
                "relation {}\n".format(pw_id_1, pw_id_2, str(', '.join(map(str, col_names))),
                                       str(relations[rl_id].relation_name)))
            print(str(diff))
        return diff

    @staticmethod
    def difference_both_ways(relations: list, dfs: list, rl_id: int, pw_id_1: int, pw_id_2: int,
                             col_names: list=None, do_print: bool=True):

        if not col_names:
            col_names = list(dfs[rl_id])[1:]

        x1 = PWEQuery.difference(relations, dfs, rl_id, pw_id_1, pw_id_2, col_names, False)
        x2 = PWEQuery.difference(relations, dfs, rl_id, pw_id_2, pw_id_1, col_names, False)

        diff = x1.append(x2, ignore_index=True)

        if do_print:
            print("Following tuples are in one of PW {} or {}, but not both, for relation {} and features {}\n".format(
                pw_id_1, pw_id_2, str(relations[rl_id].relation_name), str(', '.join(map(str, col_names)))))
            print(str(diff))
        return diff

    @staticmethod
    def redundant_column(relations: list, expected_pws: int, dfs: list, rl_id: int=0, col_names: list=None,
                               pws_to_consider: list=None, do_print: bool=True):

        if not pws_to_consider:
            pws_to_consider = [j for j in range(1, expected_pws + 1)]

        if not col_names:
            col_names = list(dfs[rl_id])[1:]

        # PW specific:
        redundant_pw_specific = []

        df = dfs[rl_id]

        for ft in col_names:
            x1 = df.groupby('pw')[ft].nunique()
            for i in pws_to_consider:
                if x1.ix[i] <= 1:
                    redundant_pw_specific.append((i, rl_id, ft))
                    if do_print:
                        print('Column ' + str(ft) + ' is redundant in relation ' + str(
                            relations[rl_id].relation_name) + ' in PW ' + str(i))

        # Across all PWs:
        redundant_across_pws = []
        x1 = df[df.pw.isin(pws_to_consider)]
        for ft in col_names:
            if x1[ft].nunique() <= 1:
                redundant_across_pws.append((pws_to_consider, rl_id, ft))
                if do_print:
                    print('Column ' + str(ft) + ' is redundant in relation ' + str(
                        relations[rl_id].relation_name) + ' for PWs ' + str(', '.join(map(str, pws_to_consider))))

        return redundant_pw_specific, redundant_across_pws

    @staticmethod
    def unique_tuples(relations: list, expected_pws: int, dfs: list, rl_id: int=0, col_names: list=None,
                      pws_to_consider: list=None, do_print: bool=True):

        if not pws_to_consider:
            pws_to_consider = [j for j in range(1, expected_pws + 1)]

        if not col_names:
            col_names = list(dfs[rl_id])[1:]

        relevant_tuples, freqs = PWEQuery.freq(relations, expected_pws, dfs, rl_id, col_names, [], pws_to_consider,
                                               False)
        unique_tuples = []
        df = dfs[rl_id]

        for i, f in enumerate(freqs):
            if f == 1:
                expr = ''
                for k in range(len(col_names) - 1):
                    expr += str(col_names[k]) + ' == ' + "'" + str(relevant_tuples.ix[i][k]) + "'" + ' and '
                expr += str(col_names[-1]) + ' == ' + "'" + str(relevant_tuples.ix[i][-1]) + "'"
                expr += ' and pw in [' + str(', '.join(map(str, pws_to_consider))) + ']'
                s3 = df.query(expr)
                s3 = s3.reset_index(drop=True)
                unique_pw = s3.ix[0]['pw']

                unique_tuples.append((relevant_tuples.ix[i], unique_pw))

                if do_print:
                    print('The unique tuple ', tuple(relevant_tuples.ix[i]), 'occurs only in PW', unique_pw)

        return unique_tuples



#!/usr/bin/env python3
from PW_explorer.pwe_helper import rel_id_from_rel_name
import copy
from collections import defaultdict


def get_colors_list(df, id):
    idx = list(map(int, list(df[df.pw == id].x1)))
    clrs = list(df[df.pw == id].x2)

    ordered_clrs = [x for _, x in sorted(zip(idx, clrs))]
    return ordered_clrs


# get the underlying pattern in the form of a list of ints
def get_pattern(sequence):
    pattern = []
    pattern_dict = {}
    count_unique = 0

    for element in sequence:

        if element in pattern_dict:
            pattern.append(pattern_dict[element])
        else:
            pattern.append(count_unique)
            pattern_dict[element] = count_unique
            count_unique += 1

    return pattern, count_unique


def compare_patterns(p1, p2, check_cyclical_equivalence=True):
    if len(p1) != len(p2):
        return False

    if p1 == p2:
        return True

    if not check_cyclical_equivalence:
        return False

    n = len(p2)
    t2 = copy.copy(p2)

    for i in range(n - 1):
        temp = t2.pop()
        t2.insert(0, temp)
        new_pattern, _ = get_pattern(t2)
        if new_pattern == p1:
            return True

    return False


def dist_helper(s1, s2):
    pattern_s1, unique_s1 = get_pattern(s1)
    pattern_s2, unique_s2 = get_pattern(s2)

    # print s1, s2

    if unique_s1 != unique_s2:
        return 1

    dist_ = 0 if compare_patterns(pattern_s1, pattern_s2) else 1
    # print dist_

    return dist_


def dist(pw_id_1, pw_id_2, **kwargs):

    kwargs = defaultdict(lambda: None, kwargs)
    dfs = kwargs['dfs']
    relations = kwargs['relations']
    if dfs is None or relations is None:
        print("None objects passed in.")
        return -1

    if pw_id_1 == pw_id_2:
        return 0

    col_rel_id = rel_id_from_rel_name('col_2', relations)
    df = dfs[col_rel_id]

    s1 = get_colors_list(df, pw_id_1)
    s2 = get_colors_list(df, pw_id_2)

    return dist_helper(s1, s2)


#### DOESN't WORK COMPLETELY. NEED TO GENERATE NEW PATTERN EVERY TIME. NO SHORTCUTS ########

# check if two patterns are the same (cyclically equivalent)
# def compare_patterns(s1, s2):

# 	if s1 == s2:
# 		return True

# 	return cyclically_equivalent(s1, s2)

# check if two lists are cyclically equivalent
# def cyclically_equivalent(s1,s2):

# 	n, i, j = len(s1), 0, 0
# 	if n != len(s2):
# 		return False

# 	while i < n and j < n:
# 		k = 1
# 		while k <= n and s1[(i+k) % n] == s2[(j+k) % n]:
# 			k += 1
# 		if k > n:
# 			return True
# 		if s1[(i+k) % n] > s2[(j+k) % n]:
# 			i += k
# 		else:
# 			j += k
# 	return False


# Basically given u and v, check if u is a sublist of vv (v followed by another copy of v)
# def alternate_cyclically_equivalent(s1, s2):

# 	if len(s1) != len(s2):
# 		return False

# 	u = s1.copy()
# 	v = s2.copy()
# 	v.extend(v)

# 	return is_sublist(v, u)

# check if s is a sublist of l
# def is_sublist(l, s):

# 	is_a_subset = False
# 	if s == []:
# 		is_a_subset = True
# 	elif s == l:
# 		is_a_subset = True
# 	elif len(s) > len(l):
# 		is_a_subset = False
# 	else:
# 		for i in range(len(l)):
# 			if l[i] == s[0]:
# 				n = 1
# 				while (n < len(s)) and (l[i+n] == s[n]):
# 					n += 1
# 				if n == len(s):
# 					is_a_subset = True
# 					break

# 	return is_a_subset



import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

with open('requirements.txt', 'r') as fh:
    requirements = fh.read().split('\n')

setuptools.setup(
    name="PW_explorer",
    version="0.0.8",
    author="Sahil Gupta",
    author_email="",
    description="An Extensible Possible World Explorer for Answer Set Programming",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/idaks/PW-explorer",
    packages=setuptools.find_packages(),
    install_requires=requirements,
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: Apache Software License",
        "Operating System :: OS Independent",
    ],
    scripts=['PW_explorer/PWE_CLI_Scripts/pwe_run_clingo', 'PW_explorer/PWE_CLI_Scripts/pwe_load_worlds',
             'PW_explorer/PWE_CLI_Scripts/pwe_complexity_calc', 'PW_explorer/PWE_CLI_Scripts/pwe_dist_calc',
             'PW_explorer/PWE_CLI_Scripts/pwe_export', 'PW_explorer/PWE_CLI_Scripts/pwe_query',
             'PW_explorer/PWE_CLI_Scripts/pwe_visualize',
             ]
)


#!/usr/bin/env python3

import pandas as pd
import numpy as np
from .pwe_query import PWEQuery


def sym_diff_dist(pw_id_1, pw_id_2, relations, dfs, pws, rls_to_use: list=None):

    if pw_id_1 == pw_id_2:
        return 0

    if not rls_to_use:
        rls_to_use = [i for i in range(len(relations))]

    dist = 0
    for rl_id in rls_to_use:
        redundant_cols = PWEQuery.redundant_column(relations=relations, expected_pws=len(pws),  dfs=dfs, rl_id=rl_id,
                                                   pws_to_consider=[pw_id_1, pw_id_2], do_print=False)[0]
        cols_to_consider = set(list(dfs[rl_id])[1:])
        for t in redundant_cols:
            if t in cols_to_consider:
                cols_to_consider.remove(t[2])
        cols_to_consider = list(cols_to_consider)

        wt1 = 1  # TBD
        k1 = 1  # TBD

        x1 = PWEQuery.difference_both_ways(relations=relations, dfs=dfs,  rl_id=rl_id, pw_id_1=pw_id_1, pw_id_2=pw_id_2,
                                           col_names=cols_to_consider, do_print=False)
        dist += wt1 * len(x1) ** k1 if x1 is not None else 0

    return dist


def euler_overlap_diff_dist(pw_id_1, pw_id_2, rl_id, col_name, dfs, pws, relations):

    x1 = PWEQuery.freq(relations=relations, expected_pws=len(pws), dfs=dfs, rl_id=rl_id, col_names=[col_name],
                       values=['"><"'], pws_to_consider=[pw_id_1], do_print=False)[1][0]
    x2 = PWEQuery.freq(relations=relations, expected_pws=len(pws), dfs=dfs, rl_id=rl_id, col_names=[col_name],
                       values=['"><"'], pws_to_consider=[pw_id_2], do_print=False)[1][0]
    return abs(x1 - x2)



#!/usr/bin/env python3

import pandas as pd
import numpy as np
import os
import sqlite3
from .pwe_helper import mkdir_p


def get_sqlite_schema(dfs, relations):

    # this approach will take constant time since there is just one row in the exported database.
    TEST_DB_LOCATION = os.path.abspath("test.db")
    conn_t = sqlite3.connect(TEST_DB_LOCATION)
    for i, df in enumerate(dfs):
        t = df.ix[0:0]
        t.to_sql(str(relations[i].relation_name), conn_t, if_exists='replace')
    schema_q = conn_t.execute("SELECT * FROM sqlite_master WHERE type='table' ORDER BY name;")
    schemas = []
    for row in schema_q.fetchall():
        schemas.append(row[4])
    for i, df in enumerate(dfs):
        conn_t.execute('DROP TABLE ' + str(relations[i].relation_name))
    conn_t.commit()
    conn_t.close()
    os.remove(TEST_DB_LOCATION)
    return schemas


def export_to_sqlite_db(export_loc, dfs, relations, db_name):

    mkdir_p(export_loc)
    sql_db_loc = os.path.join(export_loc, '{}.{}'.format(str(db_name), 'db'))
    sql_conn = sqlite3.connect(sql_db_loc)

    for i, df in enumerate(dfs):
        rel_name = str(relations[i].relation_name)
        df.to_sql(rel_name, sql_conn, if_exists='replace')

    sql_conn.commit()
    sql_conn.close()

    return True


def export(export_format, export_loc, dfs, relations):

    if export_format not in ['csv', 'h5', 'msg', 'pkl']:
        return False

    mkdir_p(export_loc)

    for i, df in enumerate(dfs):
        rel_name = str(relations[i].relation_name)
        fname = os.path.join(export_loc, '{}.{}'.format(rel_name, export_format))
        if export_format == 'csv':
            df.to_csv(fname)
        elif export_format == 'h5':
            df.to_hdf(fname, mode='w', key='df')
        elif export_format == 'msg':
            df.to_msgpack(fname)
        elif export_format == 'pkl':
            df.to_pickle(fname)

    return True
